use std::{env, ffi, ptr};

use x11::xlib::{self, BadGC};

use super::error;

pub struct Display {
    ptr: *mut xlib::Display,
    name: String,
}

impl Display {
    /// Call XOpenDisplay to open a connection to the X Server.
    /// If `display_name` is `None`, the value of the `DISPLAY` environment variable will be used.
    pub fn open(display_name: Option<String>) -> anyhow::Result<Self> {
        let name = ffi::CString::new(if let Some(name) = display_name {
            name
        } else {
            env::var("DISPLAY")?
        })?;

        let name_ptr = name.as_bytes().as_ptr();

        // try to open display and get either display pointer or null
        let display_ptr = unsafe { xlib::XOpenDisplay(name_ptr as *const i8) };

        // if display is null, return an error, otherwise return instance successfully
        if display_ptr == ptr::null_mut::<xlib::_XDisplay>() {
            Err(error::XError::OpenDisplayError(name.into_string()?).into())
        } else {
            Ok(Self {
                ptr: display_ptr,
                name: name.into_string()?,
            })
        }
    }

    /// Calls XFlush to flush the output buffer.
    pub fn flush(&self) {
        unsafe {
            xlib::XFlush(self.ptr)
        }
    }

    // TODO: Figure out how to properly handle errors
    /// Calls XSync to flush the output buffer and then wait until all events have been received and processed
    /// by the server. 
    /// The `discard` parameter specifies, whether to discard all events in the queue.
    pub fn sync(&self, discard: bool) {
        unsafe {
            xlib::XSync(self.ptr, discard.into());
        }
    }

}

impl Drop for Display {
    fn drop(&mut self) {
        if unsafe { xlib::XCloseDisplay(self.ptr) } == BadGC.into() {
            eprintln!("BadGC Error when closing display '{}'.", self.name);
        };
    }
}
